"""Content classes for rendering objects and markdown to HTML."""

import base64
import re
from abc import ABC, abstractmethod
from collections import namedtuple
from io import BytesIO, StringIO
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Set, Tuple, TypeVar, Union
from uuid import uuid4

import markdown as md

from esparto import _INSTALLED_MODULES
from esparto._options import options
from esparto.design.base import AbstractContent, AbstractLayout, Child
from esparto.design.layout import Row
from esparto.publish.output import nb_display

if "PIL" in _INSTALLED_MODULES:
    from PIL.Image import Image as PILImage  # type: ignore

if "pandas" in _INSTALLED_MODULES:
    from pandas import DataFrame  # type: ignore

if "matplotlib" in _INSTALLED_MODULES:
    from matplotlib.figure import Figure as MplFigure  # type: ignore

if "bokeh" in _INSTALLED_MODULES:
    from bokeh.embed import components  # type: ignore
    from bokeh.models.layouts import LayoutDOM as BokehObject  # type: ignore

if "plotly" in _INSTALLED_MODULES:
    from plotly.graph_objs._figure import Figure as PlotlyFigure  # type: ignore
    from plotly.io import to_html as plotly_to_html  # type: ignore


T = TypeVar("T", bound="Content")


class Content(AbstractContent, ABC):
    """Template for Content elements.

    Attributes:
        content (Any): Item to be included in the page - should match the encompassing Content class.

    """

    content: Any
    _dependencies: Set[str]

    @abstractmethod
    def to_html(self, **kwargs: bool) -> str:
        """Convert content to HTML string.

        Returns:
            str: HTML string.

        """
        raise NotImplementedError

    def display(self) -> None:
        """Display rendered content in a Jupyter Notebook cell."""
        nb_display(self)

    def __add__(self, other: Child) -> "Row":
        from esparto.design.layout import Row

        return Row(children=[self, other])

    def __iter__(self) -> Iterator["Content"]:
        return iter([self])

    def __len__(self) -> int:
        return len(list(self.content))

    def _repr_html_(self) -> None:
        nb_display(self)

    def __str__(self) -> str:
        return getattr(self, "title", None) or self.__class__.__name__

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            if hasattr(self.content, "__iter__") and hasattr(other.content, "__iter__"):
                return all(x == y for x, y in zip(self.content, other.content))
            return bool(self.content == other.content)
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


class RawHTML(Content):
    """Raw HTML content.

    Args:
        html (str): HTML string.

    """

    _dependencies: Set[Any] = set("")
    content: str

    def __init__(self, html: str) -> None:

        if not isinstance(html, str):
            raise TypeError(r"HTML must be str")

        self.content = html

    def to_html(self, **kwargs: bool) -> str:
        return self.content


class Markdown(Content):
    """Markdown text content.

    Args:
        text (str): Markdown text to be added to document.

    """

    _dependencies = {"bootstrap"}

    def __init__(self, text: str) -> None:

        if not isinstance(text, str):
            raise TypeError(r"text must be str")

        self.content: str = text

    def to_html(self, **kwargs: bool) -> str:
        html = md.markdown(self.content, extensions=["extra", "smarty"])
        html = f"{html}\n"
        html = f"<div class='es-markdown'>\n{html}\n</div>"
        return html


class Image(Content):
    """Image content.

    Can be read from a filepath, PIL.Image object, or from bytes..

    Args:
        image (str, Path, PIL.Image, BytesIO): Image data.
        caption (str): Image caption (default = None)
        alt_text (str): Alternative text. (default = None)
        scale (float): Scale image by proportion. (default = None)
        set_width (int): Set width in pixels. (default = None)
        set_height (int): Set height in pixels. (default = None)

    """

    _dependencies = {"bootstrap"}

    def __init__(
        self,
        image: Union[str, Path, "PILImage", BytesIO],
        caption: Optional[str] = "",
        alt_text: Optional[str] = "Image",
        scale: Optional[float] = None,
        set_width: Optional[int] = None,
        set_height: Optional[int] = None,
    ):

        valid_types: Tuple[Any, ...]

        if "PIL" in _INSTALLED_MODULES:
            valid_types = (str, Path, PILImage, BytesIO)
        else:
            valid_types = (str, Path, BytesIO)

        if not isinstance(image, (valid_types)):
            raise TypeError(r"`image` must be one of {}".format(valid_types))

        self.content = image
        self.alt_text = alt_text
        self.caption = caption
        self._scale = scale
        self._width = set_width
        self._height = set_height

    def set_width(self, width: int) -> None:
        """Set width of image.

        Args:
            width (int): New width in pixels.

        """
        self._width = width

    def set_height(self, height: int) -> None:
        """Set height of image.

        Args:
            height (int): New height in pixels.

        """
        self._height = height

    def rescale(self, scale: float) -> None:
        """Resize the image by a scaling factor.

        Args:
            scale (float): Scaling ratio.

        """
        self._scale = scale

    def to_html(self, **kwargs: bool) -> str:
        image_bytes = image_to_bytes(self.content)
        image_encoded = bytes_to_base64(image_bytes)

        width = f"min({self._width}, 100%)" if self._width else "auto"
        height = f"min({self._height}, 100%)" if self._height else "auto"
        scale = f"transform: scale({self._scale});" if self._scale else ""

        html = (
            "<figure class='es-figure'>"
            "<img class='img-fluid figure-img rounded es-image' "
            f"style='width: {width}; height: {height}; {scale}' "
            f"alt='{self.alt_text}' "
            f"src='data:image/png;base64,{image_encoded}'>"
        )

        if self.caption:
            html += f"<figcaption class='figure-caption'>{self.caption}</figcaption>"

        html += "</figure>"

        return html


class DataFramePd(Content):
    """Pandas DataFrame to be converted to table.

    Args:
        df (pd.DataFrame): A Pandas DataFrame
        index (bool): If True, render the DataFrame index. (default = True)
        col_space (str, int): Minimum column width in CSS units. Use int for pixels. (default = 0)

    Attributes:
        css_classes (List[str]): CSS classes applied to the HTML output.

    """

    _dependencies = {"bootstrap"}

    def __init__(
        self, df: "DataFrame", index: bool = True, col_space: Union[int, str] = 0
    ):

        if not isinstance(df, DataFrame):
            raise TypeError(r"df must be Pandas DataFrame")

        self.content: "DataFrame" = df
        self.index = index
        self.col_space = col_space
        self.css_classes = [
            "table",
            "table-hover",
        ]

    def to_html(self, **kwargs: bool) -> str:
        html: str = self.content.to_html(
            index=self.index,
            border=0,
            col_space=self.col_space,
            classes=self.css_classes,
        )
        html = f"<div class='table-responsive es-table'>{html}</div>"
        return html


class FigureMpl(Content):
    """Matplotlib figure.

    Args:
        figure (plt.Figure): A Matplotlib figure.
        output_format (str): 'svg' or 'png'. (default = None)
        pdf_figsize (tuple, float): Set figure size for PDF output. (default = None)
            Accepts a tuple of (height, width) or a float to use as scale factor.

    """

    _dependencies = {"bootstrap"}

    def __init__(
        self,
        figure: "MplFigure",
        output_format: Optional[str] = None,
        pdf_figsize: Optional[Union[Tuple[int, int], float]] = None,
    ) -> None:

        if not isinstance(figure, MplFigure):
            raise TypeError(r"figure must be a Matplotlib Figure")

        self.content: MplFigure = figure
        self.output_format = output_format or options.matplotlib.html_output_format
        self.pdf_figsize = pdf_figsize or options.matplotlib.pdf_figsize

        self._original_figsize = figure.get_size_inches()

    def to_html(self, **kwargs: bool) -> str:

        if kwargs.get("notebook_mode"):
            output_format = options.matplotlib.notebook_format
        else:
            output_format = self.output_format

        if kwargs.get("pdf_mode") and self.pdf_figsize:
            if isinstance(self.pdf_figsize, float):
                figsize = self.pdf_figsize * self._original_figsize
            else:
                figsize = self.pdf_figsize
            self.content.set_size_inches(*figsize)

        if output_format == "svg":

            string_buffer = StringIO()
            self.content.savefig(string_buffer, format="svg")
            string_buffer.seek(0)
            xml = string_buffer.read()

            dpi = 96
            fig_width, fig_height = (
                int(val * dpi) for val in self.content.get_size_inches()
            )

            if kwargs.get("pdf_mode"):
                xml = responsive_svg_mpl(
                    xml, width=int(fig_width), height=int(fig_height)
                )
                temp_file = Path(options._pdf_temp_dir) / f"{uuid4()}.svg"
                temp_file.write_text(xml)
                inner = (
                    "<object type='image/svg+xml' width='100%' height='100%' "
                    f"data='{temp_file.name}'></object>\n"
                )
            else:
                xml = responsive_svg_mpl(xml)
                inner = xml

            html = (
                f"<div class='es-matplotlib-figure' style='width: min({fig_width}px, 100%);'>"
                f"{inner}\n</div>\n"
            )

            # Reset figsize in case it was changed
            self.content.set_size_inches(*self._original_figsize)

            return html

        # If not svg:
        bytes_buffer = BytesIO()
        self.content.savefig(bytes_buffer, format="png")
        bytes_buffer.seek(0)
        return Image(bytes_buffer).to_html()


class FigureBokeh(Content):
    """Bokeh object to be rendered as an interactive plot.

    Args:
        figure (bokeh.layouts.LayoutDOM): A Bokeh object.
        layout_attributes (dict): Attributes set on `figure`. (default = None)

    """

    _dependencies = {"bokeh"}

    def __init__(
        self,
        figure: "BokehObject",
        layout_attributes: Optional[Dict[Any, Any]] = None,
    ):
        if not issubclass(type(figure), BokehObject):
            raise TypeError(r"figure must be a Bokeh object")

        self.content: BokehObject = figure
        self.layout_attributes = layout_attributes or options.bokeh.layout_attributes

    def to_html(self, **kwargs: bool) -> str:

        if self.layout_attributes:
            for key, value in self.layout_attributes.items():
                setattr(self.content, key, value)

        # Bokeh to PDF is experimental and untested
        if kwargs.get("pdf_mode"):  # pragma: no cover
            from bokeh.io import export_svg  # type: ignore

            temp_file = Path(options._pdf_temp_dir) / f"{uuid4()}.svg"
            export_svg(self.content, filename=str(temp_file))
            html = f"<img src='{temp_file.name}' width='100%' height='auto'>\n"
            return html

        html, js = components(self.content)

        # Remove outer <div> tag so we can give our own attributes
        html = remove_outer_div(html)

        fig_width = self.content.properties_with_values().get("width", 1000)

        return (
            "<div class='es-bokeh-figure' "
            f"style='width: min({fig_width}px, 100%);'>"
            f"\n{html}\n{js}\n</div>"
        )


class FigurePlotly(Content):
    """Plotly figure to be rendered as an interactive plot.

    Args:
        figure (plotly.graph_objs._figure.Figure): A Plotly figure.
        layout_args (dict): Args passed to `figure.update_layout()`. (default = None)

    """

    _dependencies = {"plotly"}

    def __init__(
        self,
        figure: "PlotlyFigure",
        layout_args: Optional[Dict[Any, Any]] = None,
    ):

        if not isinstance(figure, PlotlyFigure):
            raise TypeError(r"figure must be a Plotly Figure")

        self.layout_args = layout_args or options.plotly.layout_args

        self.content: PlotlyFigure = figure
        self._original_layout = figure.layout

    def to_html(self, **kwargs: bool) -> str:

        if self.layout_args:
            self.content.update_layout(**self.layout_args)

        # Default width is 700, default height is 450
        fig_width: int = getattr(self.content, "width", 700)

        if kwargs.get("pdf_mode"):
            temp_file = Path(options._pdf_temp_dir) / f"{uuid4()}.svg"
            self.content.write_image(str(temp_file))
            inner = f"<img src='{temp_file.name}' width='100%' height='auto'>"

        else:
            inner = plotly_to_html(
                self.content, include_plotlyjs=False, full_html=False
            )
            # Remove outer <div> tag so we can give our own attributes.
            inner = remove_outer_div(inner)

        html = f"<div class='es-plotly-figure' style='width: min({fig_width}px, 100%);'>{inner}\n</div>"

        # Reset layout in case it was changed
        self.content.update_layout(self._original_layout)

        return html


def remove_outer_div(html: str) -> str:
    """Remove outer <div> tags."""
    html = html.replace("<div>", "", 1)
    html = "".join(html.rsplit("</div>", 1))
    return html


def image_to_bytes(image: Union[str, Path, BytesIO, "PILImage"]) -> BytesIO:
    """Convert `image` to bytes.

    Args:
        image (Union[str, Path, BytesIO, PIL.Image]): image object.

    Raises:
        TypeError: image type not recognised.

    Returns:
        BytesIO: image as bytes object.

    """
    if isinstance(image, BytesIO):
        return image
    elif "PIL" in _INSTALLED_MODULES and isinstance(image, PILImage):
        return BytesIO(image.tobytes())
    elif isinstance(image, (str, Path)):
        return BytesIO(Path(image).read_bytes())
    else:
        raise TypeError(type(image))


def bytes_to_base64(bytes: BytesIO) -> str:
    """
    Convert an image from bytes to base64 representation.

    Args:
        image (BytesIO): image bytes object.

    Returns:
        str: image encoded as a base64 utf-8 string.

    """
    return base64.b64encode(bytes.getvalue()).decode("utf-8")


def table_of_contents(
    object: AbstractLayout, max_depth: Optional[int] = None, numbered: bool = True
) -> "Markdown":
    """Produce table of contents for a Layout object.

    Args:
        object (Layout): Target object for TOC.
        max_depth (int): Maximum depth of returned TOC.
        numbered (bool): If True TOC items are numbered.
            If False, bulletpoints are used.

    """
    from esparto.design.content import Markdown

    max_depth = max_depth or 99

    TOCItem = namedtuple("TOCItem", "title, level, id")

    def get_toc_items(parent: AbstractLayout) -> List[TOCItem]:
        def find_ids(parent: Any, level: int, acc: List[TOCItem]) -> List[TOCItem]:
            if hasattr(parent, "get_title_identifier") and parent.title:
                acc.append(TOCItem(parent.title, level, parent.get_title_identifier()))
                level += 1
            if hasattr(parent, "children"):
                for child in parent.children:
                    find_ids(child, level, acc)
            else:
                return acc
            return acc

        acc_new = find_ids(parent, 0, [])
        return acc_new

    toc_items = get_toc_items(object)

    tab = "\t"
    marker = "1." if numbered else "*"
    markdown_list = [
        f"{(item.level - 1) * tab} {marker} [{item.title}](#{item.id})"
        for item in toc_items
        if item.level > 0 and item.level <= max_depth
    ]
    markdown_str = "\n".join(markdown_list)

    return Markdown(markdown_str)


def responsive_svg_mpl(
    source: str, width: Optional[int] = None, height: Optional[int] = None
) -> str:
    """Make SVG element responsive."""

    regex_w = r"width=\S*"
    regex_h = r"height=\S*"

    width_ = f"width='{width}px'" if width else ""
    height_ = f"height='{height}px'" if height else ""

    source = re.sub(regex_w, width_, source, count=1)
    source = re.sub(regex_h, height_, source, count=1)

    # Preserve aspect ratio of SVG
    old_str = r"<svg"
    new_str = '<svg class="svg-content-mpl" preserveAspectRatio="xMinYMin meet" '
    source = source.replace(old_str, new_str, 1)

    return source



"""Content dependency management."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Set

from esparto import _INSTALLED_MODULES
from esparto._options import options


@dataclass
class ContentDependency:
    name: str
    cdn: str
    inline: str
    target: str


@dataclass
class ResolvedDeps:
    head: List[str] = field(default_factory=list)
    tail: List[str] = field(default_factory=list)


class ContentDependencyDict(dict):  # type: ignore
    def __add__(self, item: ContentDependency) -> "ContentDependencyDict":
        super().__setitem__(item.name, item)
        return self


JS_DEPS = {"bokeh", "plotly"}


def lazy_content_dependency_dict() -> ContentDependencyDict:
    bootstrap_inline = Path(options.bootstrap_css).read_text()
    bootstrap_inline = f"<style>\n{bootstrap_inline}\n</style>"

    content_dependency_dict = ContentDependencyDict()
    content_dependency_dict += ContentDependency(
        "bootstrap", options.bootstrap_cdn, bootstrap_inline, "head"
    )

    if "bokeh" in _INSTALLED_MODULES:
        import bokeh.resources as bk_resources  # type: ignore

        bokeh_cdn = bk_resources.CDN.render_js()
        bokeh_inline = bk_resources.INLINE.render_js()

        content_dependency_dict += ContentDependency(
            "bokeh", bokeh_cdn, bokeh_inline, "tail"
        )

    if "plotly" in _INSTALLED_MODULES:
        from plotly import offline as plotly_offline  # type: ignore

        plotly_version = "latest"
        plotly_cdn = f"<script src='https://cdn.plot.ly/plotly-{plotly_version}.min.js'></script>"
        plotly_inline = plotly_offline.get_plotlyjs()
        plotly_inline = f"<script>\n{plotly_inline}\n</script>"

        content_dependency_dict += ContentDependency(
            "plotly", plotly_cdn, plotly_inline, "head"
        )

    return content_dependency_dict


def resolve_deps(required_deps: Set[str], source: Optional[str]) -> ResolvedDeps:
    resolved_deps = ResolvedDeps()

    if source not in {"cdn", "inline"}:
        raise ValueError("Dependency source must be one of {'cdn', 'inline'}")

    source = options.dependency_source

    for dep in required_deps:
        dep_details: ContentDependency = lazy_content_dependency_dict()[dep]
        getattr(resolved_deps, dep_details.target).append(getattr(dep_details, source))

    return resolved_deps



"""Abstract design classes to help decoupling of domain from implementation."""

from abc import ABC
from typing import Any, List, Set, TypeVar, Union

T = TypeVar("T", bound="AbstractLayout")

Child = Union["AbstractLayout", "AbstractContent", Any]


class AbstractLayout(ABC):
    """Class Template for Layout elements.

    Layout class hierarchy:
        `Page -> Section -> Row -> Column -> Content`

    Attributes:
        title (str): Object title. Used as a title within the page and as a key value.
        children (list): Child items defining the page layout and content.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    # ------------------------------------------------------------------------+
    #                              Public Methods                             |
    # ------------------------------------------------------------------------+

    def display(self) -> None:
        """Render content in a Notebook environment."""
        raise NotImplementedError

    def get_identifier(self) -> str:
        """Get the HTML element ID for the current object."""
        raise NotImplementedError

    def get_title_identifier(self) -> str:
        """Get the HTML element ID for the current object title."""
        raise NotImplementedError

    def set_children(self, other: Union[List[Child], Child]) -> None:
        """Set children as `other`."""
        raise NotImplementedError

    def to_html(self, **kwargs: bool) -> str:
        """Render object as HTML string.

        Returns:
            html (str): HTML string.

        """
        raise NotImplementedError

    def tree(self) -> None:
        """Display page tree."""
        raise NotImplementedError


class AbstractContent(ABC):
    """Template for Content elements.

    Attributes:
        content (Any): Item to be included in the page - should match the encompassing Content class.

    """

    content: Any
    _dependencies: Set[str]

    def to_html(self, **kwargs: bool) -> str:
        """Convert content to HTML string.

        Returns:
            str: HTML string.

        """
        raise NotImplementedError

    def display(self) -> None:
        """Display rendered content in a Jupyter Notebook cell."""
        raise NotImplementedError



"""Esparto configuration options."""

import collections.abc
import copy
import pprint
import traceback
from contextlib import ContextDecorator
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Any, Dict, Mapping, Optional, Tuple, Type, Union

import yaml

from esparto import _MODULE_PATH


class ConfigMixin(object):
    _options_source: str

    def _to_dict(self) -> Dict[str, Any]:
        return public_dict(self.__dict__)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        string = f"{pprint.pformat(self._to_dict(), sort_dicts=False)}"
        if hasattr(self, "_options_source"):
            string += (
                f"\nSource: {self._options_source}" if self._options_source else ""
            )
        return string


@dataclass(repr=False)
class MatplotlibOptions(yaml.YAMLObject, ConfigMixin):
    """Options for Matplotlib output.

    Attributes:
        html_output_format (str):
            How plots are rendered in HTML: 'png' or 'svg'.
        notebook_format (str):
            How plots are rendered in Jupyter Notebooks: 'png' or 'svg'.
        pdf_figsize (tuple or int):
            Specify size of Matplotlib figures in PDF output.
            An integer tuple can be passed as: (height, width).
            A float can be passed as a scaling factor.

    """

    yaml_loader = yaml.SafeLoader
    yaml_tag = "!MatplotlibOptions"

    html_output_format: str = "svg"
    notebook_format: str = "svg"
    pdf_figsize: Optional[Union[Tuple[int, int], float]] = 1.0


@dataclass(repr=False)
class PlotlyOptions(yaml.YAMLObject, ConfigMixin):
    """Options for Plotly output.

    Attributes:
        layout_args (dict):
            Arguments passed to `figure.update_layout()` at rendering time.

    """

    yaml_loader = yaml.SafeLoader
    yaml_tag = "!PlotlyOptions"

    layout_args: Dict[str, Any] = field(default_factory=lambda: {})


@dataclass(repr=False)
class BokehOptions(yaml.YAMLObject, ConfigMixin):
    """Options for Bokeh output.

    Attributes:
        layout_attributes (dict):
            Bokeh layout object attributes to set at rendering time.

    """

    yaml_loader = yaml.SafeLoader
    yaml_tag = "!BokehOptions"

    layout_attributes: Dict[str, Any] = field(
        default_factory=lambda: {"sizing_mode": "scale_width"}
    )


@dataclass(repr=False)
class OutputOptions(yaml.YAMLObject, ConfigMixin):
    """Options for configuring page rendering and output.

    Config options will automatically be loaded if a yaml file is found at
    either `./esparto-config.yaml` or `~/esparto-data/esparto-config.yaml`.

    Attributes:
        dependency_source (str):
            How dependencies should be provisioned: 'cdn' or 'inline'.
        bootstrap_cdn (str):
            Link to Bootstrap CDN. Used if dependency source is 'cdn'.
        bootstrap_css (str):
            Path to Bootstrap CSS file. Used if dependency source is 'inline'.
        esparto_css (str):
            Path to additional CSS file with esparto specific styles.
        jinja_template (str):
            Path to Jinja HTML page template.

        matplotlib: Additional config options for Matplotlib.
        plotly: Additional config options for Plotly.
        bokeh: Additional config options for Bokeh.

    """

    yaml_loader = yaml.SafeLoader
    yaml_tag = "!OutputOptions"

    dependency_source: str = "cdn"
    bootstrap_cdn: str = (
        "<link rel='stylesheet' "
        "href='https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/5.1.3/css/bootstrap.min.css' "
        "integrity='sha512-GQGU0fMMi238uA+a/bdWJfpUGKUkBdgfFdgBm72SUQ6BeyWjoY/ton0tEjH+OSH9iP4Dfh+7HM0I9f5eR0L/4w==' "
        "crossorigin='anonymous' referrerpolicy='no-referrer'>"
    )
    bootstrap_css: str = str(_MODULE_PATH / "resources/css/bootstrap.min.css")
    esparto_css: str = str(_MODULE_PATH / "resources/css/esparto.css")
    jinja_template: str = str(_MODULE_PATH / "resources/jinja/base.html.jinja")

    matplotlib: MatplotlibOptions = MatplotlibOptions()
    bokeh: BokehOptions = BokehOptions()
    plotly: PlotlyOptions = PlotlyOptions()

    _pdf_temp_dir: str = TemporaryDirectory().name

    _options_source: str = ""

    def save(self, path: Union[str, Path] = "./esparto-config.yaml") -> None:
        """Save config to yaml file at `path`."""
        Path(path).write_text(self._to_yaml_str())

    @classmethod
    def load(cls, path: Union[str, Path]) -> "OutputOptions":
        """Load config from yaml file at `path`."""
        yaml_str = Path(path).read_text()
        opts: OutputOptions = yaml.safe_load(yaml_str)
        opts._options_source = str(path)
        return opts

    def _to_yaml_str(self) -> str:
        self_copy = copy.copy(self)
        del self_copy._options_source
        return str(yaml.dump(self_copy, default_flow_style=False, sort_keys=False))

    @classmethod
    def _autoload(cls) -> "OutputOptions":
        config_paths = [
            Path("./esparto-config.yaml"),
            Path.home() / "esparto-data/esparto-config.yaml",
        ]

        for p in config_paths:
            if p.is_file():
                opts = cls.load(p)
                opts._options_source = str(p)
                print("esparto config loaded from:", p)
                return opts
        return cls()


options = OutputOptions._autoload()


def update_recursive(
    source_dict: Dict[Any, Any], update_map: Mapping[Any, Any]
) -> Dict[Any, Any]:
    """Recursively update nested dictionaries.
    https://stackoverflow.com/a/3233356/8065696
    """
    for k, v in update_map.items():
        if isinstance(v, collections.abc.Mapping):
            source_dict[k] = update_recursive(source_dict.get(k, {}), v)
        else:
            source_dict[k] = v
    return source_dict


class options_context(ContextDecorator):
    def __init__(self, page_options: OutputOptions):
        self.page_options = page_options
        self.default_options = copy.copy(options)

    def __enter__(self) -> None:
        update_recursive(options.__dict__, self.page_options.__dict__)

    def __exit__(
        self, exc_type: Type[BaseException], exc_value: BaseException, tb: TracebackType
    ) -> None:
        if exc_type is not None:  # pragma: no cover
            traceback.print_exception(exc_type, exc_value, tb)
        update_recursive(options.__dict__, self.default_options.__dict__)


def resolve_config_option(config_option: str, value: Optional[str]) -> Any:
    if value is None:
        return getattr(options, config_option)
    else:
        return value


def public_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    """Remove keys starting with '_' from dict."""
    return {
        k: v for k, v in d.items() if not (isinstance(k, str) and k.startswith("_"))
    }



"""Functions that render and save documents."""

import time
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from bs4 import BeautifulSoup, Tag  # type: ignore
from jinja2 import Template

from esparto import _INSTALLED_MODULES
from esparto._options import options, resolve_config_option
from esparto.design.base import AbstractContent, AbstractLayout
from esparto.publish.contentdeps import resolve_deps

if TYPE_CHECKING:
    from esparto.design.layout import Page


def publish_html(
    page: "Page",
    filepath: Optional[str] = "./esparto-doc.html",
    return_html: bool = False,
    dependency_source: Optional[str] = None,
    esparto_css: Optional[str] = None,
    jinja_template: Optional[str] = None,
    **kwargs: bool,
) -> Optional[str]:
    """Save page to HTML.

    Args:
      page (Page): A Page object.
      filepath (str): Filepath to write to.
      return_html (bool): Returns HTML string if True.
      dependency_source (str): One of 'cdn' or 'inline' (default = None).
      esparto_css (str): Path to CSS stylesheet. (default = None).
      jinja_template (str): Path to Jinja template. (default = None).
      **kwargs (Dict[str, Any]): Arguments passed to `page.to_html()`.

    Returns:
      str: HTML string if return_html is True.

    """

    required_deps = page._required_dependencies()
    dependency_source = dependency_source or options.dependency_source
    resolved_deps = resolve_deps(required_deps, source=dependency_source)

    esparto_css = Path(resolve_config_option("esparto_css", esparto_css)).read_text()

    page_html = page.to_html(**kwargs)
    jinja_template_object = Template(
        Path(resolve_config_option("jinja_template", jinja_template)).read_text()
    )
    html_rendered: str = jinja_template_object.render(
        navbrand=page.navbrand,
        doc_title=page.title,
        esparto_css=esparto_css,
        content=page_html,
        head_deps=resolved_deps.head,
        tail_deps=resolved_deps.tail,
    )
    html_rendered = prettify_html(html_rendered)
    html_rendered = relocate_scripts(html_rendered)

    if filepath:
        Path(filepath).write_text(html_rendered)

    if return_html:
        return html_rendered
    return None


def publish_pdf(
    page: "Page", filepath: str = "./esparto-doc.pdf", return_html: bool = False
) -> Optional[str]:
    """Save page to PDF.

    Args:
      page (Layout): A Page object.
      filepath (str): Filepath to write to.
      return_html (bool): Returns HTML string if True.

    Returns:
      str: HTML string if return_html is True.

    """
    if "weasyprint" not in _INSTALLED_MODULES:
        raise ModuleNotFoundError("Install weasyprint for PDF support")
    import weasyprint as wp  # type: ignore

    temp_dir = Path(options._pdf_temp_dir)
    temp_dir.mkdir(parents=True, exist_ok=True)

    html_rendered = publish_html(
        page=page,
        filepath=None,
        return_html=True,
        dependency_source="inline",
        pdf_mode=True,
    )
    pdf_doc = wp.HTML(string=html_rendered, base_url=options._pdf_temp_dir).render()
    pdf_doc.metadata.title = page.title
    pdf_doc.write_pdf(filepath)

    for f in temp_dir.iterdir():
        f.unlink()
    temp_dir.rmdir()

    html_prettified = prettify_html(html_rendered)

    if return_html:
        return html_prettified
    return None


def nb_display(
    item: Union["AbstractLayout", "AbstractContent"],
    return_html: bool = False,
    dependency_source: Optional[str] = None,
) -> Optional[str]:
    """Display Layout or Content to Jupyter Notebook cell.

    Args:
      item (Layout, Content): A Layout or Content item.
      return_html (bool): Returns HTML string if True.
      dependency_source (str): One of 'cdn', 'inline', or 'esparto.options'.

    Returns:
      str: HTML string if return_html is True.

    """
    from IPython.display import HTML, display  # type: ignore

    from esparto.design.layout import Layout

    if isinstance(item, Layout):
        required_deps = item._required_dependencies()
    else:
        required_deps = getattr(item, "_dependencies", set())

    dependency_source = dependency_source or options.dependency_source
    resolved_deps = resolve_deps(required_deps, source=dependency_source)
    esparto_css = Path(options.esparto_css).read_text()
    head_deps = "\n".join(resolved_deps.head)
    tail_deps = "\n".join(resolved_deps.tail)
    html = item.to_html(notebook_mode=True)
    html_rendered = (
        f"<div class='container' style='width: 100%; height: 100%;'>\n{html}\n</div>\n"
    )
    html_rendered += f"<style>\n{esparto_css}\n</style>\n"

    html_rendered = (
        f"<!doctype html>\n<html>\n<head>{head_deps}</head>\n"
        f"<body>\n{html_rendered}\n{tail_deps}\n</body>\n</html>\n"
    )
    html_rendered = relocate_scripts(html_rendered)
    print()

    # This allows time to download plotly.js from the CDN - otherwise cell can render empty
    if "plotly" in required_deps and dependency_source == "cdn":
        display(HTML(f"<head>\n{head_deps}\n</head>\n"), metadata=dict(isolated=True))
        time.sleep(2)

    # Temporary solution to prevent Jupyter Notebook cell fully collapsing before content renders
    if "bokeh" in required_deps:
        extra_css = "<style>.container { min-height: 30em !important; }</style>"
    else:
        extra_css = ""

    display(HTML(extra_css + html_rendered), metadata=dict(isolated=True))
    print()

    if return_html:
        return html_rendered
    return None


def prettify_html(html: Optional[str]) -> str:
    """Prettify HTML."""
    html = html or ""
    html = str(BeautifulSoup(html, features="html.parser").prettify())

    return html


def relocate_scripts(html: str) -> str:
    """Move all JavaScript in page body to end of section."""
    soup = BeautifulSoup(html, "html.parser")
    body = soup.find("body")

    if isinstance(body, Tag):
        script_list = body.find_all("script")
        for script in script_list:
            body.insert(len(body), script)

    html = str(soup)

    return html



"""Layout classes for defining page appearance and structure."""

import copy
import re
from abc import ABC
from pprint import pformat
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

import bs4  # type: ignore

from esparto._options import OutputOptions, options, options_context
from esparto.design.base import AbstractLayout, Child
from esparto.publish.output import nb_display, publish_html, publish_pdf

T = TypeVar("T", bound="Layout")


class Layout(AbstractLayout, ABC):
    """Class Template for Layout elements.

    Layout class hierarchy:
        `Page -> Section -> Row -> Column -> Content`

    Attributes:
        title (str): Object title. Used as a title within the page and as a key value.
        children (list): Child items defining the page layout and content.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    # ------------------------------------------------------------------------+
    #                              Magic Methods                              |
    # ------------------------------------------------------------------------+

    title: Optional[str]
    children: List[Child] = []

    title_html_tag: str
    title_classes: List[str]
    title_styles: Dict[str, Any]

    body_html_tag: str
    body_classes: List[str]
    body_styles: Dict[str, Any]

    @property
    def _default_id(self) -> str:
        return f"es-{type(self).__name__}".lower()

    @property
    def _parent_class(self) -> Type["Layout"]:
        raise NotImplementedError

    @property
    def _child_class(self) -> Type["Layout"]:
        raise NotImplementedError

    _dependencies = {"bootstrap"}

    @property
    def _child_ids(self) -> Dict[str, str]:
        """Return existing child IDs or a new dict."""
        try:
            super().__getattribute__("__child_ids")
        except AttributeError:
            super().__setattr__("__child_ids", {})
        child_ids: Dict[str, str] = super().__getattribute__("__child_ids")
        return child_ids

    def __init__(
        self,
        title: Optional[str] = None,
        children: Union[List[Child], Child] = None,
        title_classes: Optional[List[str]] = None,
        title_styles: Optional[Dict[str, Any]] = None,
        body_classes: Optional[List[str]] = None,
        body_styles: Optional[Dict[str, Any]] = None,
    ):
        self.title = title
        children = children or []
        self.set_children(children)

        self.__post_init__()

        title_classes = title_classes or []
        title_styles = title_styles or {}
        body_classes = body_classes or []
        body_styles = body_styles or {}

        self.title_classes += title_classes
        self.title_styles.update(title_styles)

        self.body_classes += body_classes
        self.body_styles.update(body_styles)

    def __post_init__(self) -> None:
        raise NotImplementedError

    def __iter__(self) -> Iterator["Layout"]:
        return iter([self])

    def __repr__(self) -> str:
        return self._tree()

    def _repr_html_(self) -> None:
        self.display()

    def __str__(self) -> str:
        return self._tree()

    def __add__(self: T, other: Child) -> T:
        new = copy.copy(self)
        new.children = self.children + [*self._smart_wrap(other)]
        return new

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.title == other.title
                and len(self.children) == len(other.children)
                and all((x == y for x, y in zip(self.children, other.children)))
            )
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __getattribute__(self, key: str) -> Any:
        child_id = super().__getattribute__("_child_ids").get(key)
        if child_id:
            return self.__getitem__(child_id)
        return super().__getattribute__(key)

    def __setattr__(self, key: str, value: Any) -> None:
        child_id = super().__getattribute__("_child_ids").get(key)
        if child_id:
            self.__setitem__(child_id, value)
        else:
            super().__setattr__(key, value)

    def __delattr__(self, key: str) -> None:
        child_id = super().__getattribute__("_child_ids").get(key)
        if child_id:
            self.__delitem__(child_id)
        else:
            super().__delattr__(key)

    def __getitem__(self, key: Union[str, int]) -> Any:
        if isinstance(key, str):
            indexes = get_matching_titles(key, self.children)
            if len(indexes) and key:
                return self.children[indexes[0]]
            value = self._child_class(title=key)
            self.children.append(value)
            if key:
                self._add_child_id(key)
            return self.children[-1]

        elif isinstance(key, int):
            if key < len(self.children):
                return self.children[key]
            value = self._child_class()
            self.children.append(value)
            return self.children[-1]

        raise KeyError(key)

    def __setitem__(self, key: Union[str, int], value: Any) -> None:
        value = copy.copy(value)
        title = (
            getattr(value, "title", None) if issubclass(type(value), Layout) else None
        )
        if not isinstance(value, self._child_class):
            if issubclass(self._child_class, Column):
                value = self._child_class(title=title, children=[value])
            else:
                value = self._smart_wrap(value)
                value = value[0]
        if isinstance(key, str):
            if key:
                value.title = title or key
                indexes = get_matching_titles(key, self.children)
                if indexes:
                    self.children[indexes[0]] = value
                else:
                    self.children.append(value)
                self._add_child_id(value.title)
            else:
                self.children.append(value)
            return
        elif isinstance(key, int):
            if key < len(self.children):
                value.title = title or getattr(self.children[key], "title", None)
                self.children[key] = value
                return
            self.children.append(value)
            return

        raise KeyError(key)

    def __delitem__(self, key: Union[int, str]) -> None:
        if isinstance(key, str):
            indexes = get_matching_titles(key, self.children)
            if len(indexes):
                self._remove_child_id(key)
                del self.children[indexes[0]]
                return None
        elif isinstance(key, int) and key < len(self.children):
            child_title = getattr(self.children[key], "title", None)
            if child_title:
                self._remove_child_id(child_title)
            del self.children[key]
            return None
        raise KeyError(key)

    def __lshift__(self, other: Child) -> Child:
        self.set_children(other)
        return other

    def __rshift__(self, other: Child) -> "Layout":
        self.set_children(other)
        return self

    def __copy__(self) -> "Layout":
        attributes = vars(self)
        new = self.__class__()
        new.__dict__.update(attributes)
        new.children = [*new.children]
        return new

    # ------------------------------------------------------------------------+
    #                              Public Methods                             |
    # ------------------------------------------------------------------------+

    def display(self) -> None:
        """Render content in a Notebook environment."""
        nb_display(self)

    def get_identifier(self) -> str:
        """Get the HTML element ID for the current object."""
        return clean_attr_name(str(self.title)) if self.title else self._default_id

    def get_title_identifier(self) -> str:
        """Get the HTML element ID for the current object title."""
        return f"{self.get_identifier()}-title"

    def set_children(self, other: Union[List[Child], Child]) -> None:
        """Set children as `other`."""
        other = copy.copy(other)
        self.children = [*self._smart_wrap(other)]
        for child in self.children:
            title = getattr(child, "title", None)
            if title:
                self._add_child_id(title)

    def to_html(self, **kwargs: bool) -> str:
        """Render object as HTML string.

        Returns:
            html (str): HTML string.

        """
        children_rendered = " ".join([c.to_html(**kwargs) for c in self.children])
        title_rendered = (
            render_html(
                self.title_html_tag,
                self.title_classes,
                self.title_styles,
                self.title,
                self.get_title_identifier(),
            )
            if self.title
            else ""
        )
        html = render_html(
            self.body_html_tag,
            self.body_classes,
            self.body_styles,
            f"{title_rendered}\n{children_rendered}\n",
            self.get_identifier(),
        )
        html = bs4.BeautifulSoup(html, "html.parser").prettify()
        return html

    def tree(self) -> None:
        """Display page tree."""
        print(self._tree())

    # ------------------------------------------------------------------------+
    #                             Private Methods                             |
    # ------------------------------------------------------------------------+

    def _add_child_id(self, key: str) -> None:
        attr_name = clean_attr_name(key)
        if attr_name:
            self._child_ids[attr_name] = key
            super().__setattr__(attr_name, self[key])

    def _remove_child_id(self, key: str) -> None:
        attr_name = clean_attr_name(key)
        if attr_name in self._child_ids:
            del self._child_ids[attr_name]
            super().__delattr__(attr_name)

    def _smart_wrap(self, child_list: Union[List[Child], Child]) -> List[Child]:
        """Wrap children in a coherent class hierarchy.

        Args:
            children: Sequence of Content and / or Child items.

        Returns:
            List of Layout and Content items wrapped in a coherent class hierarchy.


        If the parent object is a Column and the item is a Content Class:
            - return child with no modification
        If the parent object is a Column and the item is not a Content Class:
            - cast the child to an appropriate Content Class if possible
            - return the child
        If the current item is wrapped and unwrapped items have been accumulated:
            - wrap the unwrapped children
            - append newly wrapped to output
            - append current child to output
        If the current child is wrapped and we have no accumulated unwrapped items:
            - append the wrapped child to output
        If the current child is a dict and the parent is a Row:
            - use the dictionary key as a title and value as content
            - wrap and append the current child to output
        If the current child is unwrapped and the parent is a Row:
            - wrap and append the current child to output
        If the current item is unwrapped and the parent is not a Row:
            - add the current child to unwrapped item accumulator
        Finally:
            - wrap any accumulated unwrapped items
            - append the final wrapped segment to output

        """
        return smart_wrap(self, child_list)

    def _recurse_children(self, idx: int) -> Dict[str, Any]:
        key = self.title or f"{type(self).__name__} {idx}"
        tree = {
            f"{key}": [
                child._recurse_children(idx)  # type: ignore
                if hasattr(child, "_recurse_children")
                else str(child)
                for idx, child in enumerate(self.children)
            ]
        }
        return tree

    def _required_dependencies(self) -> Set[str]:
        deps: Set[str] = self._dependencies

        def dep_finder(parent: Any) -> None:
            nonlocal deps
            for child in parent.children:
                deps = deps | set(getattr(child, "_dependencies", {}))
                if hasattr(child, "children"):
                    dep_finder(child)

        dep_finder(self)
        return deps

    def _tree(self) -> str:
        return pformat(self._recurse_children(idx=0))

    def _ipython_key_completions_(self) -> List[str]:  # pragma: no cover
        return [
            getattr(child, "title")
            for child in self.children
            if hasattr(child, "title")
        ]


class Page(Layout):
    """Layout class that defines a Page.

    Args:
        title (str): Used as a title within the page and as a key value.
        navbrand (str): Brand name. Displayed in the page navbar if provided.
        table_of_contents (bool, int): Add a Table of Contents to the top of page.
            Passing an `int` will define the maximum depth.
        max_width (int): Maximum page width expressed in pixels.
        output_options (es.OutputOptions): Page specific rendering and output options.
        children (list): Child items defining layout and content.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    output_options: OutputOptions = options

    def __init__(
        self,
        title: Optional[str] = None,
        navbrand: Optional[str] = "",
        table_of_contents: Union[bool, int] = False,
        max_width: int = 800,
        output_options: Optional[OutputOptions] = None,
        children: Union[List[Child], Child] = None,
        title_classes: Optional[List[str]] = None,
        title_styles: Optional[Dict[str, Any]] = None,
        body_classes: Optional[List[str]] = None,
        body_styles: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            title, children, title_classes, title_styles, body_classes, body_styles
        )
        self.navbrand = navbrand
        self.table_of_contents = table_of_contents
        self.max_width = max_width
        self.output_options = output_options or options

    def save(
        self,
        filepath: str = "./esparto-doc.html",
        return_html: bool = False,
        dependency_source: Optional[str] = None,
    ) -> Optional[str]:
        """
        Save page to HTML file.

        Note:
            Alias for `self.save_html()`.

        Args:
            filepath (str): Destination filepath.
            return_html (bool): If True, return HTML as a string.
            dependency_source (str): 'cdn' or 'inline'.

        Returns:
            html (str): Document rendered as HTML. (If `return_html` is True)

        """
        html = self.save_html(
            filepath=filepath,
            return_html=return_html,
            dependency_source=dependency_source,
        )

        if return_html:
            return html
        return None

    @options_context(output_options)
    def save_html(
        self,
        filepath: str = "./esparto-doc.html",
        return_html: bool = False,
        dependency_source: Optional[str] = None,
    ) -> Optional[str]:
        """
        Save page to HTML file.

        Args:
            filepath (str): Destination filepath.
            return_html (bool): If True, return HTML as a string.
            dependency_source (str): 'cdn' or 'inline'.

        Returns:
            html (str): Document rendered as HTML. (If `return_html` is True)

        """
        html = publish_html(
            self,
            filepath=filepath,
            return_html=return_html,
            dependency_source=dependency_source,
        )
        if return_html:
            return html
        return None

    @options_context(output_options)
    def save_pdf(
        self, filepath: str = "./esparto-doc.pdf", return_html: bool = False
    ) -> Optional[str]:
        """
        Save page to PDF file.

        Note:
            Requires `weasyprint` library.

        Args:
            filepath (str): Destination filepath.
            return_html (bool): If True, return intermediate HTML representation as a string.

        Returns:
            html (str): Document rendered as HTML. (If `return_html` is True)

        """
        html = publish_pdf(self, filepath, return_html=return_html)
        if return_html:
            return html
        return None

    @options_context(output_options)
    def to_html(self, **kwargs: bool) -> str:
        if self.table_of_contents:
            # Create a copy of the page and dynamically generate the TOC.
            # Copy is required so that TOC is not added multiple times and
            # always reflects the current content.
            from esparto.design.content import table_of_contents

            max_depth = (
                None if self.table_of_contents is True else self.table_of_contents
            )
            page_copy = copy.copy(self)
            toc = table_of_contents(page_copy, max_depth=max_depth)
            page_copy.children.insert(
                0,
                page_copy._child_class(
                    title="Contents", children=[toc], title_classes=["h4"]
                ),
            )
            page_copy.table_of_contents = False
            return page_copy.to_html(**kwargs)

        self.body_styles.update({"max-width": f"{self.max_width}px"})

        return super().to_html(**kwargs)

    def __post_init__(self) -> None:
        self.title_html_tag = "h1"
        self.title_classes = ["es-page-title"]
        self.title_styles = {}

        self.body_html_tag = "div"
        self.body_classes = ["es-page-body"]
        self.body_styles = {}

    @property
    def _parent_class(self) -> Type["Layout"]:
        return Page

    @property
    def _child_class(self) -> Type["Layout"]:
        return Section


class Section(Layout):
    """Layout class that defines a Section.

    Args:
        title (str): Used as a title within the page and as a key value.
        children (list): Child items defining layout and content.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    def __post_init__(self) -> None:
        self.title_html_tag = "h3"
        self.title_classes = ["es-section-title"]
        self.title_styles = {}

        self.body_html_tag = "div"
        self.body_classes = ["es-section-body"]
        self.body_styles = {}

    @property
    def _parent_class(self) -> Type["Layout"]:
        return Page

    @property
    def _child_class(self) -> Type["Layout"]:
        return Row


class CardSection(Section):
    """Layout class that defines a CardSection. CardSections wrap content in Cards by default.

    Args:
        title (str): Used as a title within the page and as a key value.
        children (list): Child items defining layout and content.
        cards_equal (bool): Cards in the same Row are stretched vertically if True.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    def __init__(
        self,
        title: Optional[str] = None,
        children: Union[List[Child], Child, None] = None,
        cards_equal: bool = False,
        title_classes: Optional[List[str]] = None,
        title_styles: Optional[Dict[str, Any]] = None,
        body_classes: Optional[List[str]] = None,
        body_styles: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            title=title,
            children=children,
            title_classes=title_classes,
            title_styles=title_styles,
            body_classes=body_classes,
            body_styles=body_styles,
        )

        self.cards_equal = cards_equal

    @property
    def _child_class(self) -> Type["Layout"]:
        # Attribute missing if class is not instantiated
        if hasattr(self, "cards_equal") and self.cards_equal:
            return CardRowEqual
        return CardRow


class Row(Layout):
    """Layout class that defines a Row.

    Args:
        title (str): Used as a title within the page and as a key value.
        children (list): Child items defining layout and content.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    def __post_init__(self) -> None:
        self.title_html_tag = "h5"
        self.title_classes = ["col-12", "es-row-title"]
        self.title_styles = {}

        self.body_html_tag = "div"
        self.body_classes = ["row", "es-row-body"]
        self.body_styles = {}

    @property
    def _parent_class(self) -> Type["Layout"]:
        return Section

    @property
    def _child_class(self) -> Type["Layout"]:
        return Column

    def __setitem__(self, key: Union[str, int], value: Any) -> None:
        if isinstance(value, dict):
            title, content = list(value.items())[0]
            value = self._child_class(title=title, children=[content])
        super().__setitem__(key, value)


class Column(Layout):
    """Layout class that defines a Column.

    Args:
        title (str): Used as a title within the page and as a key value.
        children (list): Child items defining layout and content.
        col_width (int): Fix column width - must be between 1 and 12.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    def __init__(
        self,
        title: Optional[str] = None,
        children: Union[List[Child], Child] = None,
        col_width: Optional[int] = None,
        title_classes: Optional[List[str]] = None,
        title_styles: Optional[Dict[str, Any]] = None,
        body_classes: Optional[List[str]] = None,
        body_styles: Optional[Dict[str, Any]] = None,
    ):
        self.title = title
        children = children or []
        self.set_children(children)
        self.col_width = col_width

        self.__post_init__()

        title_classes = title_classes or []
        title_styles = title_styles or {}
        body_classes = body_classes or []
        body_styles = body_styles or {}

        self.title_classes += title_classes
        self.title_styles.update(title_styles)

        self.body_classes += body_classes
        self.body_styles.update(body_styles)

    def __post_init__(self) -> None:
        self.title_html_tag = "h5"
        self.title_classes = ["es-column-title"]
        self.title_styles = {}

        col_class = f"col-lg-{self.col_width}" if self.col_width else "col-lg"
        self.body_html_tag = "div"
        self.body_classes = [col_class, "es-column-body"]
        self.body_styles = {}

    @property
    def _parent_class(self) -> Type["Layout"]:
        return Row

    @property
    def _child_class(self) -> Type["Layout"]:
        raise NotImplementedError


class CardRow(Row):
    """Layout class that defines a CardRow. CardRows wrap content in Cards by default.

    Args:
        title (str): Used as a title within the page and as a key value.
        children (list): Child items defining layout and content.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    @property
    def _child_class(self) -> Type["Layout"]:
        return Card


class CardRowEqual(CardRow):
    """Layout class that defines a CardRow with Cards of equal height.

    Args:
        title (str): Used as a title within the page and as a key value.
        children (list): Child items defining layout and content.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    def __post_init__(self) -> None:
        super().__post_init__()
        self.body_styles = {"align-items": "stretch"}


class Card(Column):
    """Layout class that defines a Card.

    Child items will be vertically stacked by default.
    Horizontal arrangement is achieved by nesting content inside a Row.

    Args:
        title (str): Used as a title within the page and as a key value.
        children (list): Child items defining layout and content.
        col_width (int): Fix column width - must be between 1 and 12.
        title_classes (list): Additional CSS classes to apply to title.
        title_styles (dict): Additional CSS styles to apply to title.
        body_classes (list): Additional CSS classes to apply to body.
        body_styles (dict): Additional CSS styles to apply to body.

    """

    def __init__(
        self,
        title: Optional[str] = None,
        children: Union[List[Child], Child] = None,
        col_width: int = 6,
        title_classes: Optional[List[str]] = None,
        title_styles: Optional[Dict[str, Any]] = None,
        body_classes: Optional[List[str]] = None,
        body_styles: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            title=title,
            children=children,
            col_width=col_width,
            title_classes=title_classes,
            title_styles=title_styles,
            body_classes=body_classes,
            body_styles=body_styles,
        )

    def __post_init__(self) -> None:
        self.title_html_tag = "h5"
        self.title_classes = ["card-title", "es-card-title"]
        self.title_styles = {}

        self.body_html_tag = "div"

        col_class = f"col-lg-{self.col_width}" if self.col_width else "col-lg"
        self.body_classes = [col_class, "es-card"]
        self.body_styles = {}

    def to_html(self, **kwargs: bool) -> str:
        """Render content to HTML string.

        Returns:
            html (str): HTML string.

        """
        children_rendered = " ".join([c.to_html(**kwargs) for c in self.children])
        title_rendered = (
            render_html(
                self.title_html_tag,
                self.title_classes,
                self.title_styles,
                self.title,
                self.get_title_identifier(),
            )
            if self.title
            else ""
        )
        card_body_classes = ["es-card-body"]
        card_body_styles: Dict[str, str] = {}
        html_body = render_html(
            "div",
            card_body_classes,
            card_body_styles,
            f"\n{title_rendered}\n{children_rendered}\n",
            f"{self.get_identifier()}-body",
        )
        html_full = render_html(
            self.body_html_tag,
            self.body_classes,
            self.body_styles,
            f"\n{html_body}\n",
            f"{self.get_identifier()}",
        )

        return html_full


class Spacer(Column):
    """Empty Column for making space within a Row."""


class PageBreak(Section):
    """Defines a page break when printing or saving to PDF."""

    body_id = "es-page-break"

    def __post_init__(self) -> None:
        self.title_html_tag = ""
        self.title_classes = []
        self.title_styles = {}

        self.body_html_tag = "div"
        self.body_classes = []
        self.body_styles = {}


def smart_wrap(self: Layout, child_list: Union[List[Child], Child]) -> List[Child]:
    from esparto.design.adaptors import content_adaptor

    child_list = ensure_iterable(child_list)

    if isinstance(self, Column):
        if any((isinstance(x, dict) for x in child_list)):
            raise TypeError("Invalid content passed to Column: 'dict'")
        return [content_adaptor(x) for x in child_list]

    is_row = isinstance(self, Row)
    unwrapped_acc: List[Child] = []
    output: List[Child] = []

    for child in child_list:
        is_wrapped = isinstance(child, self._child_class)

        if is_wrapped:
            if unwrapped_acc:
                wrapped_segment = self._child_class(children=unwrapped_acc)
                output.append(wrapped_segment)
                output.append(child)
                unwrapped_acc = []
            else:
                output.append(child)
        else:  # if not is_wrapped
            if is_row:
                if isinstance(child, dict):
                    title, child = list(child.items())[0]
                else:
                    title = None
                output.append(self._child_class(title=title, children=[child]))
            else:
                unwrapped_acc.append(child)

    if unwrapped_acc:
        wrapped_segment = self._child_class(children=unwrapped_acc)
        output.append(wrapped_segment)

    return output


def render_html(
    tag: str,
    classes: List[str],
    styles: Dict[str, str],
    children: str,
    identifier: Optional[str] = None,
) -> str:
    """Render HTML from provided attributes."""
    class_str = " ".join(classes) if classes else ""
    class_str = f"class='{class_str}'" if classes else ""

    style_str = "; ".join((f"{key}: {value}" for key, value in styles.items()))
    style_str = f"style='{style_str}'" if styles else ""

    id_str = f"id='{identifier}'" if identifier else ""

    rendered = " ".join((f"<{tag} {id_str} {class_str} {style_str}>").split())
    rendered += f"\n  {children}\n</{tag}>"

    return rendered


def get_index_where(
    condition: Callable[..., bool], iterable: Iterable[Any]
) -> List[int]:
    """Return index values where `condition` is `True`."""
    return [idx for idx, item in enumerate(iterable) if condition(item)]


def get_matching_titles(title: str, children: List["Child"]) -> List[int]:
    """Return child items with matching title."""
    return get_index_where(lambda x: bool(getattr(x, "title", None) == title), children)


def clean_attr_name(attr_name: str) -> str:
    """Remove invalid characters from the attribute name."""
    if not attr_name:
        return ""

    # Remove leading and trailing spaces
    attr_name = attr_name.strip().replace(" ", "_").lower()

    # Remove invalid characters
    attr_name = re.sub("[^0-9a-zA-Z_]", "", attr_name)

    # Remove leading characters until we find a letter or underscore
    attr_name = re.sub("^[^a-zA-Z_]+", "", attr_name)

    return attr_name


def ensure_iterable(something: Any) -> Iterable[Any]:
    # Convert any non-list iterators to lists
    iterable = (
        list(something) if isinstance(something, (list, tuple, set)) else [something]
    )
    # Un-nest any nested lists of children
    if len(list(iterable)) == 1 and isinstance(list(iterable)[0], (list, tuple, set)):
        iterable = list(iterable)[0]
    return iterable



import functools as ft
import mimetypes as mt
from pathlib import Path
from typing import Any, Dict, Union

from esparto import _INSTALLED_MODULES
from esparto.design.content import (
    Content,
    DataFramePd,
    FigureBokeh,
    FigureMpl,
    FigurePlotly,
    Image,
    Markdown,
)
from esparto.design.layout import Layout


@ft.singledispatch
def content_adaptor(content: Content) -> Union[Content, Layout, Dict[str, Any]]:
    """
    Wrap content in the required class. If Layout object is passed, return unchanged.

    Args:
      content (Any): Any content to be added to the document.

    Returns:
      Content: Appropriately wrapped content.

    """
    if not issubclass(type(content), Content):
        raise TypeError(f"Unsupported content type: {type(content)}")
    return content


@content_adaptor.register(str)
@content_adaptor.register(Path)
def content_adaptor_core(content: Union[str, Path]) -> Content:
    """Convert text or image to Markdown or Image content."""
    content = str(content)
    guess = mt.guess_type(content)
    if guess and isinstance(guess[0], str):
        file_type = guess[0].split("/")[0]
        if file_type == "image":
            return Image(content)
        elif file_type == "text":
            content = Path(content).read_text()
        else:
            raise TypeError(f"{content}: {file_type}")
    return Markdown(content)


@content_adaptor.register(Layout)
def content_adaptor_layout(content: Layout) -> Layout:
    """If Layout object is passed, return unchanged."""
    return content


@content_adaptor.register(dict)
def content_adaptor_dict(content: Dict[str, Any]) -> Dict[str, Any]:
    """Pass through dict of `{"title": content}`."""
    if not (len(content) == 1 and isinstance(list(content.keys())[0], str)):
        raise ValueError("Content dict must be passed as {'title': content}")
    return content


# Function only available if Pandas is installed.
if "pandas" in _INSTALLED_MODULES:
    from pandas.core.frame import DataFrame  # type: ignore

    @content_adaptor.register(DataFrame)
    def content_adaptor_df(content: DataFrame) -> DataFramePd:
        """Convert Pandas DataFrame to DataFramePD content."""
        return DataFramePd(content)


# Function only available if Matplotlib is installed.
if "matplotlib" in _INSTALLED_MODULES:
    from matplotlib.figure import Figure  # type: ignore

    @content_adaptor.register(Figure)
    def content_adaptor_mpl(content: Figure) -> FigureMpl:
        """Convert Matplotlib Figure to FigureMpl content."""
        return FigureMpl(content)


# Function only available if Bokeh is installed.
if "bokeh" in _INSTALLED_MODULES:
    from bokeh.layouts import LayoutDOM as BokehObject  # type: ignore

    @content_adaptor.register(BokehObject)
    def content_adaptor_bokeh(content: BokehObject) -> FigureBokeh:
        """Convert Bokeh Layout to FigureBokeh content."""
        return FigureBokeh(content)


# Function only available if Plotly is installed.
if "plotly" in _INSTALLED_MODULES:
    from plotly.graph_objs._figure import Figure as PlotlyFigure  # type: ignore

    @content_adaptor.register(PlotlyFigure)
    def content_adaptor_plotly(content: PlotlyFigure) -> FigurePlotly:
        """Convert Plotly Figure to FigurePlotly content."""
        return FigurePlotly(content)



"""Command line utilities for esparto."""

from argparse import SUPPRESS, ArgumentParser, _SubParsersAction
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import esparto._options as opt
from esparto import __version__

PROG = "esparto"
DESCRIPTION = "Command line utilities for esparto."
EPILOG = "Run program with no arguments for subcommand help."

parser = ArgumentParser(prog=PROG, usage=None, description=DESCRIPTION, epilog=EPILOG)
subparsers = parser.add_subparsers(dest="subcommand")

parser.add_argument("-v", "--version", action="version", version=__version__)


CliArg = Tuple[List[str], Dict[str, Any]]


def argument(*name_or_flags: str, **kwargs: Dict[str, Any]) -> CliArg:
    """Convenience function to properly format arguments to pass to the
    subcommand decorator.
    """
    return (list(name_or_flags), kwargs)


def subcommand(
    *subparser_args: CliArg, parent: _SubParsersAction = subparsers
) -> Callable[..., Any]:
    """Decorator to define a new subcommand in a sanity-preserving way.
    The function will be stored in the ``func`` variable when the parser
    parses arguments so that it can be called directly like so::
        args = cli.parse_args()
        args.func(args)
    Usage example::
        @subcommand([argument("-d", help="Enable debug mode", action="store_true")])
        def subcommand(args):
            print(args)
    Then on the command line::
        $ python cli.py subcommand -d
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        parser_ = parent.add_parser(
            func.__name__, description=func.__doc__, add_help=False, usage=SUPPRESS
        )
        for args, kwargs in subparser_args:
            parser_.add_argument(*args, **kwargs)
        parser_.set_defaults(func=func)
        return func

    return decorator


@subcommand()
def print_esparto_css(*args: Any) -> None:
    """print default esparto CSS"""
    css = Path(opt.OutputOptions.esparto_css).read_text()
    print(css)


@subcommand()
def print_bootstrap_css(*args: Any) -> None:
    """print default Bootstrap CSS"""
    css = Path(opt.OutputOptions.bootstrap_css).read_text()
    print(css)


@subcommand()
def print_jinja_template(*args: Any) -> None:
    """print default jinja template"""
    css = Path(opt.OutputOptions.jinja_template).read_text()
    print(css)


@subcommand()
def print_default_options(*args: Any) -> None:
    """print default output options"""
    print(opt.OutputOptions()._to_yaml_str())


def print_subcommand_help() -> None:
    """Print help for subcommands."""
    subparser_actions = [
        action for action in parser._actions if isinstance(action, _SubParsersAction)
    ]
    print("subcommands:")
    for subparsers_action in subparser_actions:
        left_pad_size = max(len(x) for x in subparsers_action.choices.keys()) + 2
        left_pad_size = max(left_pad_size, 22)
        for choice, subparser in subparsers_action.choices.items():
            print(f"  {choice:<{left_pad_size}}{subparser.format_help().strip()}")


def main() -> None:
    args = parser.parse_args()
    if args.subcommand is None:
        parser.print_help()
        print()
        print_subcommand_help()
    else:
        args.func(args)



# -*- coding: utf-8 -*-
"""
esparto
=======

Data driven report builder for the PyData ecosystem.

Please visit https://domvwt.github.io/esparto/ for documentation and examples.

"""

from importlib.util import find_spec as _find_spec
from pathlib import Path as _Path
from typing import Set as _Set

__author__ = """Dominic Thorn"""
__email__ = "dominic.thorn@gmail.com"
__version__ = "4.0.0"

_MODULE_PATH: _Path = _Path(__file__).parent.absolute()


_OPTIONAL_DEPENDENCIES: _Set[str] = {
    "PIL",  # Only used for type checking and conversion
    "IPython",
    "matplotlib",
    "pandas",
    "bokeh",
    "plotly",
    "weasyprint",
}

_INSTALLED_MODULES: _Set[str] = {
    x.name for x in [_find_spec(dep) for dep in _OPTIONAL_DEPENDENCIES] if x
}

from esparto._options import OutputOptions, options
from esparto.design.content import (
    DataFramePd,
    FigureBokeh,
    FigureMpl,
    FigurePlotly,
    Image,
    Markdown,
    RawHTML,
)
from esparto.design.layout import (
    Card,
    CardRow,
    CardRowEqual,
    CardSection,
    Column,
    Page,
    PageBreak,
    Row,
    Section,
    Spacer,
)
